/* File:        $Id: ArcFilesReportGenerator.java 2251 2012-02-08 13:03:03Z mss $
 * Revision:    $Revision: 2251 $
 * Author:      $Author: mss $
 * Date:        $Date: 2012-02-08 14:03:03 +0100 (Wed, 08 Feb 2012) $
 *
 * The Netarchive Suite - Software to harvest and preserve websites
 * Copyright 2004-2012 The Royal Danish Library, the Danish State and
 * University Library, the National Library of France and the Austrian
 * National Library.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
package dk.netarkivet.harvester.harvesting;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintWriter;
import java.text.MessageFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.LinkedHashMap;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import dk.netarkivet.common.exceptions.ArgumentNotValid;
import dk.netarkivet.common.exceptions.IOFailure;

/**
 * This class generate a report that lists ARC files
 * along with the opening date, closing date (if file was properly closed),
 * and size in bytes.
 *
 * Here is a sample of such a file:
 *
 * [ARCFILE] [Opened] [Closed] [Size]
 *  5-1-20100720161253-00000-bnf_test.arc.gz "2010-07-20 16:12:53.698" "2010-07-20 16:14:31.792" 162928
 *
 * The file is named "arcfiles-report.txt" and is generated by parsing the
 * "heritrix.out" file located in the crawl directory. Useful lines match the
 * following examples:
 *
 * 2010-07-20 16:12:53.698 INFO thread-14 org.archive.io.WriterPoolMember.createFile() Opened /somepath/jobs/current/high/5_1279642368951/arcs/5-1-20100720161253-00000.arc.gz.open
 *
 * and
 *
 * 2010-07-20 16:14:31.792 INFO thread-29 org.archive.io.WriterPoolMember.close() Closed /somepath/jobs/current/high/5_1279642368951/arcs/5-1-20100720161253-00000-bnf_test.arc.gz, size 162928
 *
 * In order to have such messages output to heritrix.out,
 * the "heritrix.properties" file must contain the following, uncommented line:
 *
 * org.archive.io.arc.ARCWriter.level = INFO
 *
 * Note that these strings have changed between Heritrix version 1.14.3
 * and 1.14.4, so they might change again in the future.
 *
 */
class ArcFilesReportGenerator {

    private static final Log LOG =
        LogFactory.getLog(ArcFilesReportGenerator.class);

    private static final SimpleDateFormat ISO_8601_DATE_FORMAT =
        new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.SSS'Z'");

    private static final SimpleDateFormat SOURCE_DATE_FORMAT =
        new SimpleDateFormat("yyyy-MM-dd HH:mm:ss.SSS");

    /**
     * Stores the opening date, closing date and size of an ARC file.
     */
    static class ArcFileStatus {

        private static final String NOT_AVAILABLE = "?";

        String openedDate = NOT_AVAILABLE;
        String closedDate = NOT_AVAILABLE;
        long size = 0;

        protected String getOpenedDate() {
            return openedDate;
        }

        protected void setOpenedDate(String openedDate) {
            this.openedDate = getIsoDateString(openedDate);
        }

        protected String getClosedDate() {
            return closedDate;
        }

        protected void setClosedDate(String closedDate) {
            this.closedDate = getIsoDateString(closedDate);
        }

        protected long getSize() {
            return size;
        }
        protected void setSize(long size) {
            this.size = size;
        }

        @Override
        public String toString() {
            return openedDate + " " + closedDate + " " + Long.toString(size);
        }

        private String getIsoDateString(String dateStr) {
            try {
                return ISO_8601_DATE_FORMAT.format(
                        SOURCE_DATE_FORMAT.parse(dateStr));
            } catch (ParseException e) {
                return NOT_AVAILABLE;
            }
        }

    }

    /**
     * Format used to parse and extract values from  lines of heritrix.out
     * pertaining to an ARC file opening.
     */
    public static final MessageFormat ARC_OPEN_FORMAT =
        new MessageFormat(
                "{0} INFO thread-{1} "
                + "org.archive.io.WriterPoolMember.createFile() Opened "
                + "{2}.open");

    /**
     * Format used to parse and extract values from lines of heritrix.out
     * pertaining to an ARC file closing.
     */
    public static final MessageFormat ARC_CLOSE_FORMAT =
        new MessageFormat(
                "{0} INFO thread-{1} "
                + "org.archive.io.WriterPoolMember.close() Closed {2}"
                + ", size {3}");

    /**
     * The name of the report file. It will be generated in the crawl directory.
     */
    public static final String REPORT_FILE_NAME = "arcfiles-report.txt";

    /**
     * The Heritrix crawl directory.
     */
    private File crawlDir;

    /**
     * Builds a ARC files report generator, given the Heritrix crawl directory.
     * @param crawlDir the Heritrix crawl directory.
     */
    ArcFilesReportGenerator(File crawlDir) {
        this.crawlDir = crawlDir;
    }

    /**
     * Parses heritrix.out and generates the ARC files report.
     * @return the generated report file.
     */
    protected File generateReport() {

        Map<String, ArcFileStatus> reportContents = parseHeritrixOut();

        File arcFilesReport = new File(crawlDir, REPORT_FILE_NAME);

        try {
            boolean created = arcFilesReport.createNewFile();
            if (!created) {
                throw new IOException("Unable to create '" 
                        + arcFilesReport.getAbsolutePath() + "'."); 
            }
            PrintWriter out = new PrintWriter(arcFilesReport);

            out.println("[ARCFILE] [Opened] [Closed] [Size]");

            for (Map.Entry<String, ArcFilesReportGenerator.ArcFileStatus> entry
                    : reportContents.entrySet()) {
                String arcFileName = entry.getKey();
                ArcFileStatus afs = entry.getValue();
                out.println(arcFileName + " " + afs.toString());
            }

            out.close();
        } catch (IOException e) {
            throw new IOFailure("Failed to create " + REPORT_FILE_NAME, e);
        }

        return arcFilesReport;
    }

    /**
     * Parses the heritrix.out file and maps to every found ARC file an
     * {@link ArcFileStatus} instance.
     * @return the map of found ARC files, and related ArcFileStatus
     */
    protected Map<String, ArcFileStatus> parseHeritrixOut() {

        Map<String, ArcFileStatus> arcFiles =
            new LinkedHashMap<String, ArcFileStatus>();

        try {
            BufferedReader heritrixOut = new BufferedReader(
                    new FileReader(new File(crawlDir, "heritrix.out")));

            String line = null;
            while ((line = heritrixOut.readLine()) != null) {

                try {
                    Object[] params = ARC_OPEN_FORMAT.parse(line);

                    String openedDate = (String) params[0];
                    String arcFileName = new File((String) params[2]).getName();

                    ArcFileStatus afs = new ArcFileStatus();
                    afs.setOpenedDate(openedDate);

                    arcFiles.put(arcFileName, afs);
                } catch (ParseException e) {
                    // NOP, that's not the line we're looking for.
                }

                try {
                    Object[] params = ARC_CLOSE_FORMAT.parse(line);

                    String closedDate = (String) params[0];
                    String arcFileName = new File((String) params[2]).getName();
                    Long size = Long.parseLong((String) params[3]);

                    ArcFileStatus afs = arcFiles.get(arcFileName);
                    if (afs == null) {
                        throw new ArgumentNotValid("ARC file " + arcFileName 
                                + " has no previous Opened record!");
                    }

                    afs.setClosedDate(closedDate);
                    afs.setSize(size);

                } catch (ParseException e) {
                    // NOP, that's not the line we're looking for.
                }

            }
            heritrixOut.close();
        } catch (IOException e) {
            LOG.error(e);
            return arcFiles;
        }

        return arcFiles;
    }
}
